@@ -20,6 +20,12 @@ along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 MA 02111-1307, USA. */
 
+#ifndef _ECM_GMP_H
+#define _ECM_GMP_H 1
+
+#include "config.h"
+#include <gmp.h>
+
 #ifndef alloca
 #ifdef __GNUC__
 # define alloca __builtin_alloca
@@ -111,9 +117,10 @@ MA 02111-1307, USA. */
   } while (0)
 #endif
 
-#ifdef HAVE_NATIVE_mpn_add_nc
+#ifdef HAVE___GMPN_ADD_NC
 #ifndef __gmpn_add_nc
-mp_limb_t __gmpn_add_nc (mp_ptr, mp_srcptr, mp_srcptr, mp_size_t, mp_limb_t);
+__GMP_DECLSPEC mp_limb_t __gmpn_add_nc (mp_ptr, mp_srcptr, mp_srcptr,
+    mp_size_t, mp_limb_t);
 #endif
 #endif
 
@@ -133,3 +140,5 @@ mp_size_t mpn_fft_next_size (mp_size_t, int);
 void    mpn_mul_fft (mp_ptr, mp_size_t, mp_srcptr, mp_size_t, mp_srcptr,
                      mp_size_t, int);
 #endif
+
+#endif /* _ECM_GMP_H */
