@@ -20,31 +20,19 @@
   MA 02111-1307, USA.
 */
 
-#include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
-#include "gmp.h"
-#include "ecm.h"
 #include "ecm-impl.h"
 
-#if WANT_ASSERT
-#include <assert.h>
-#define ASSERT(expr)   assert (expr)
-#else
-#define ASSERT(expr)   do {} while (0)
-#endif
-
-#define min(a,b) (((a)<(b))?(a):(b))
-
-/* R_i <- q_i * S, 0 <= i < n, where q_i are large positive integers, S 
-   is a point on an elliptic curve. Uses max(bits in q_i) modular 
-   inversions (one less if max(q_i) is a power of 2).
-   Needs up to n+2 cells in T.
-   Returns factor found, not found, error.
+/* R_i <- q_i * S, 0 <= i < n, where q_i are large integers, S is a point on
+   an elliptic curve. Uses max(bits in q_i) modular inversions (one less if 
+   max(q_i) is a power of 2). Needs up to n+2 cells in T.
+   Returns whether factor was found or not found, factor goes into p. 
+   No error can occur.
 */
 
 static int
-multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n, 
+multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, const unsigned int n, 
               mpmod_t modulus, mpres_t u, mpres_t v, mpres_t *T,
               unsigned long *tot_muls, unsigned long *tot_gcds)
 {
@@ -56,6 +44,8 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
   int youpi = ECM_NO_FACTOR_FOUND;
   mpz_t flag; /* Used as bit field, keeps track of which R[i] contain partial results */
   point s;    /* 2^t * S */
+  mpz_t signs; /* Used as bit field, i-th bit is set iff q[i]<0 */
+  mpz_t __dummy; /* used for local computations */
 
   if (n == 0)
     return ECM_NO_FACTOR_FOUND;
@@ -71,7 +61,8 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
       return ECM_NO_FACTOR_FOUND;
     }
   
-  mpz_init (flag);
+  MPZ_INIT2 (flag, n);
+  MPZ_INIT2 (signs, n);
   mpres_init (s.x, modulus);
   mpres_init (s.y, modulus);
   mpres_set (s.x, S->x, modulus);
@@ -82,12 +73,13 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
   maxbit = 0;
   for (i = 0; i < n; i++)
     {
+      /* We'll first compute positive multiples and change signs later */
       if (mpz_sgn (q[i]) < 0)
         {
-          outputf (OUTPUT_ERROR, "multiplyW2n: multiplicand q[%d] < 0, negatives not supported\n", i);
-	  youpi = ECM_ERROR;
-	  goto clear_w2;
+          mpz_setbit (signs, i);;
+          mpz_neg (q[i], q[i]);
         }
+
       /* Multiplier == 0? Then set result to neutral element */
       if (mpz_sgn (q[i]) == 0)
         {
@@ -174,11 +166,13 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
 #ifdef WANT_ASSERT
                 mpres_sub (u, s.x, R[i].x, modulus);
                 mpres_mul (u, u, T[l], modulus);
-                mpres_get_z (p, u, modulus);
-                mpz_mod (p, p, modulus->orig_modulus);
-                if (mpz_cmp_ui (p, 1) != 0) 
+		mpz_init(__dummy);
+                mpres_get_z (__dummy, u, modulus);
+                mpz_mod (__dummy, __dummy, modulus->orig_modulus);
+                if (mpz_cmp_ui (__dummy, 1) != 0) 
                   outputf (OUTPUT_ERROR, "Error, (s.x - R[%d].x) * T[%d] == "
-                           "%Zd\n", i, l, p);
+                           "%Zd\n", i, l, __dummy);
+		mpz_clear(__dummy);
 #endif
                 
                 mpres_sub (u, s.y, R[i].y, modulus);   /* U    = y2 - y1 */
@@ -206,11 +200,13 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
 #ifdef WANT_ASSERT
           mpres_add (u, s.y, s.y, modulus);
           mpres_mul (u, u, T[k], modulus);
-          mpres_get_z (p, u, modulus);
-          mpz_mod (p, p, modulus->orig_modulus);
-          if (mpz_cmp_ui (p, 1) != 0)
+	  mpz_init(__dummy);
+          mpres_get_z (__dummy, u, modulus);
+          mpz_mod (__dummy, __dummy, modulus->orig_modulus);
+          if (mpz_cmp_ui (__dummy, 1) != 0)
             outputf (OUTPUT_ERROR, "Error, at t==%d, 2*s.y / (2*s.y) == %Zd\n", 
-                     t, p);
+                     t, __dummy);
+	  mpz_clear(__dummy);
 #endif          
 
                                                /* 1/(2*s.y) is in T[k] */
@@ -229,7 +225,6 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
         }
     }
 
- clear_w2:
   mpres_clear (s.y, modulus);
   mpres_clear (s.x, modulus);
   mpz_clear (flag);
@@ -239,6 +234,15 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
   if (tot_gcds != NULL)
     *tot_gcds += gcds;
   
+  /* Now take inverse points (negative y-coordinate) where q[i] was < 0 */
+  for (i = 0; i < n; i++)
+    if (mpz_tstbit (signs, i))
+      {
+        mpz_neg (R[i].y, R[i].y);
+        mpz_neg (q[i], q[i]);
+      }
+
+  mpz_clear (signs);
   return youpi;
 }
 
@@ -251,7 +255,8 @@ multiplyW2n (mpz_t p, point *R, curve *S, mpz_t *q, unsigned int n,
      for (j=0;j<n;j++)
          (x[j+(n+1)*i] : y[j+(n+1)*i]) += (x[j+1+(n+1)*i] : y[j+1+(n+1)*i])
 
-   Uses one inversion and 6*n*m-3 multiplications for n*m > 0
+   Uses one inversion and 6*n*m-3 multiplications for n*m > 0.
+   Processes neutral (zero), identical and negative points correctly.
 
    Return factor found or not (no error can occur here).
 */
@@ -284,15 +289,22 @@ addWnm (mpz_t p, point *X, curve *S, mpmod_t modulus, unsigned int m,
 
         if (mpres_is_zero (T[k], modulus))  /* If both x-cordinates are identical */
           {
-            /* Are the points identical? */
+            /* Are the points identical? Compare y coordinates: */
             mpres_sub (T[k], X2->y, X1->y, modulus);
             if (mpres_is_zero (T[k], modulus))
               {
                 /* Yes, we need to double. Schedule 2*X[...].y */
                 mpres_add (T[k], X1->y, X1->y, modulus); 
               }
-            else
-              continue; /* No, they are inverses. Nothing tbd here */
+            else /* No, they are inverses. Nothing tbd here */
+              {
+#ifdef WANT_ASSERT
+                /* Check that the y coordinates are mutual negatives */
+                mpres_add (T[k], X2->y, X1->y, modulus);
+                ASSERT (mpres_is_zero (T[k], modulus));
+#endif
+                continue; 
+              }
           }
 
         if (k > 0)
@@ -418,13 +430,116 @@ addWnm (mpz_t p, point *X, curve *S, mpmod_t modulus, unsigned int m,
    or an error occurred.
 */
 
-int
-ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2, 
-            unsigned int dF, curve *s, int S, mpmod_t modulus)
+#ifdef MONT_ROOTS
+static int
+ecm_rootsF_Mont (listz_t F, root_params_t *root_params, 
+		 unsigned long dF, curve *s, mpmod_t modulus)
 {
-  unsigned int i;
+  unsigned long i, n;
+  unsigned long muls = 0;
+  long st;
+  mpres_t x6, z6, xn_6, zn_6, xn, zn; /* 6*d2*X, (n-6)*d2*X and n*d2*X, resp. */
+  mpres_t b, u, v, w;
+  mpz_t t;
+
+  if (dF == 0)
+    return ECM_NO_FACTOR_FOUND;
+
+  st = cputime ();
+
+  MPZ_INIT (t);
+
+  mpres_init (b, modulus);
+  mpres_add_ui (b, s->A, 2, modulus);
+  mpres_div_2exp (b, b, 2, modulus);
+
+  /* We init n to 7, so n-6 == 1 */
+  mpres_init (xn_6, modulus);
+  mpres_init (zn_6, modulus);
+  mpres_set (xn_6, s->x, modulus);
+  mpres_set (zn_6, s->y, modulus);
+  mpz_set_ui (t, root_params->d2);
+  ecm_mul (xn_6, zn_6, t, modulus, b);
+  /* (xn_6::zn_6) = d2*X */
+
+  mpres_init (x6, modulus);
+  mpres_init (z6, modulus);
+  mpres_set (x6, xn_6, modulus);
+  mpres_set (z6, zn_6, modulus);
+  mpz_set_ui (t, 6);
+  ecm_mul (x6, z6, t, modulus, b);
+  /* (x6::z6) = 6*d2*X */
+
+  mpres_init (xn, modulus);
+  mpres_init (zn, modulus);
+  mpres_set (xn, xn_6, modulus);
+  mpres_set (zn, zn_6, modulus);
+  mpz_set_ui (t, 7);
+  ecm_mul (xn, zn, t, modulus, b);
+  /* xn :: zn = 7*d2*X */
+
+  outputf (OUTPUT_TRACE, "ecm_rootsF_Mont: storing d2*X in F[0]\n");
+  mpres_get_z (F[0], xn_6, modulus);
+  /* F[0] = (d2*X)_x */
+
+  mpres_init (u, modulus);
+  mpres_init (v, modulus);
+  mpres_init (w, modulus);
+
+  n = 7;
+
+  for (i = 1; i < dF; )
+    {
+      if (gcd (n, root_params->d1) == 1)
+	{
+	  outputf (OUTPUT_TRACE, 
+		   "ecm_rootsF_Mont: storing d2*%lu*X in F[%lu]\n",
+		   n, i);
+	  mpres_get_z (F[i++], xn, modulus);
+	  /* Normalize coordinate to zn=1 */
+	  mpres_get_z (t, zn, modulus);
+	  mpz_invert (t, t, modulus->orig_modulus);
+	  mpz_mul (t, t, F[i++]);
+	  mpz_mod (F[i++], t, modulus->orig_modulus);
+	}
+      
+      add3 (xn_6, zn_6, xn, zn, x6, z6, xn_6, zn_6, 
+	    modulus, u, v, w);
+      mpres_swap (xn_6, xn, modulus);
+      mpres_swap (zn_6, zn, modulus);
+      n += 6; 
+      /* Now (xn::zn) = j*d2*X and (xn_6::zn_6) = (j-6)*d2*X again */
+      muls += 6;
+    }
+
+  outputf (OUTPUT_VERBOSE, "Computing roots of F took %ldms",
+	   elltime (st, cputime ()));
+  outputf (OUTPUT_DEVVERBOSE, ", and %ld muls", muls);
+  outputf (OUTPUT_VERBOSE, "\n");
+
+  mpres_clear (x6, modulus);
+  mpres_clear (z6, modulus);
+  mpres_clear (xn_6, modulus);
+  mpres_clear (zn_6, modulus);
+  mpres_clear (xn, modulus);
+  mpres_clear (zn, modulus);
+  mpres_clear (b, modulus);
+  mpres_clear (u, modulus);
+  mpres_clear (v, modulus);
+  mpres_clear (w, modulus);
+  mpz_clear (t);
+
+  return ECM_NO_FACTOR_FOUND;
+}
+#endif
+
+int
+ecm_rootsF (mpz_t f, listz_t F, root_params_t *root_params, 
+            unsigned long dF, curve *s, mpmod_t modulus)
+{
+  unsigned long i;
   unsigned long muls = 0, gcds = 0;
-  unsigned int st;
+  long st;
   int youpi = ECM_NO_FACTOR_FOUND;
   listz_t coeffs;
   ecm_roots_state state;
@@ -435,8 +550,14 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
 
   st = cputime ();
 
+#ifdef MONT_ROOTS
+  if (root_params->S == 1)
+    return ecm_rootsF_Mont (F, root_params, dF, s, modulus);
+#endif
+
   /* Relative cost of point add during init and computing roots assumed =1 */
-  init_roots_state (&state, S, d1, d2, 1.0);
+  init_roots_state (&state, root_params->S, root_params->d1, root_params->d2, 
+                    1.0);
 
   outputf (OUTPUT_DEVVERBOSE,
 	   "ecm_rootsF: state: nr = %d, dsieve = %d, size_fd = %d, S = %d, "
@@ -444,9 +565,9 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
 	   state.S, state.dickson_a);
 
   /* Init finite differences tables */
-  mpz_init (t); /* t = 0 */
-  coeffs = init_progression_coeffs (t, state.dsieve, d2, 1, 6, state.S, 
-                                    state.dickson_a);
+  MPZ_INIT (t); /* t = 0 */
+  coeffs = init_progression_coeffs (t, state.dsieve, root_params->d2, 1, 6, 
+                                    state.S, state.dickson_a);
   mpz_clear (t);
 
   if (coeffs == NULL) /* error */
@@ -474,8 +595,11 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
   for (i = 0; i < state.size_fd; i++)
     {
       outputf (OUTPUT_TRACE, "ecm_rootsF: coeffs[%d] = %Zd\n", i, coeffs[i]);
+      MEMORY_TAG;
       mpres_init (state.fd[i].x, modulus);
+      MEMORY_TAG;
       mpres_init (state.fd[i].y, modulus);
+      MEMORY_UNTAG;
     }
 
   state.T = (mpres_t *) malloc ((state.size_fd + 4) * sizeof (mpres_t));
@@ -485,7 +609,11 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
       goto ecm_rootsF_clearfdi;
     }
   for (i = 0 ; i < state.size_fd + 4; i++)
-    mpres_init (state.T[i], modulus);
+    {
+      MEMORY_TAG;
+      mpres_init (state.T[i], modulus);
+      MEMORY_UNTAG;
+    }
 
   /* Multiply fd[] = s * coeffs[] */
 
@@ -512,7 +640,7 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
     {
       unsigned int st1 = cputime ();
       outputf (OUTPUT_VERBOSE,
-	       "Initializing tables of differences for F took %ums",
+	       "Initializing tables of differences for F took %ldms",
 	       elltime (st, st1));
       outputf (OUTPUT_DEVVERBOSE, ", %lu muls and %lu extgcds", muls, gcds);
       outputf (OUTPUT_VERBOSE, "\n");
@@ -542,7 +670,7 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
             }
           
           /* Is this a j value where we want Dickson(j * d2) * X as a root? */
-          if (gcd (state.rsieve, d1) == 1) 
+          if (gcd (state.rsieve, root_params->d1) == 1) 
             mpres_get_z (F[i++], state.fd[state.next * (state.S + 1)].x, 
                          modulus);
 
@@ -568,7 +696,7 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
   if (youpi)
     return youpi; /* error or factor found */
   
-  outputf (OUTPUT_VERBOSE, "Computing roots of F took %ums",
+  outputf (OUTPUT_VERBOSE, "Computing roots of F took %ldms",
 	   elltime (st, cputime ()));
   outputf (OUTPUT_DEVVERBOSE, ", %ld muls and %ld extgcds", muls, gcds);
   outputf (OUTPUT_VERBOSE, "\n");
@@ -586,11 +714,121 @@ ecm_rootsF (mpz_t f, listz_t F, unsigned int d1, unsigned int d2,
    If a factor is found during the initialisation, NULL is returned and the
    factor in f. If an error occurred, NULL is returned and f is -1.
 */
-ecm_roots_state *
-ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
-                 unsigned int dF, unsigned int blocks, int S, mpmod_t modulus)
+
+#ifdef MONT_ROOTS
+static ecm_roots_state *
+ecm_rootsG_Mont_init (curve *X, root_params_t *root_params, 
+		      unsigned long dF, unsigned long blocks, mpmod_t modulus)
 {
-  unsigned int k, lenT, phid2;
+  unsigned long i;
+  unsigned long muls;
+  ecm_roots_state *state;
+  long st;
+  mpz_t t;
+
+  outputf (OUTPUT_DEVVERBOSE,
+           "ecm_rootsG_Mont_init: dF = %lu, k = %lu, root_params: d1 = %lu, "
+           "d2 = %lu, S = %i, i0 = %Zd\n",
+           dF, blocks, root_params->d1, root_params->d2, root_params->S,
+           root_params->i0);
+
+  st = cputime ();
+  
+  state = (ecm_roots_state *) malloc (sizeof (ecm_roots_state));
+  if (state == NULL)
+    return NULL;
+    
+  state->size_fd = 3;
+  state->nr = 1;
+  state->next = 0;
+  state->S = 1;
+  state->dsieve = root_params->d2;
+  state->rsieve = mpz_tdiv_ui (root_params->i0, root_params->d2);
+  state->dickson_a = 0;
+  /* we'll need to know which roots are {0, 1, 2}*d2*X in ecm_rootsG() */
+  if (mpz_fits_slong_p (root_params->i0))
+    state->i0 = mpz_get_si (root_params->i0);
+  else
+    state->i0 = 2;
+#ifdef MONT_ROOTS
+  state->form = EC_MONTGOMERY_FORM;
+#endif
+  state->fd = malloc (state->size_fd * sizeof (point));
+  if (state->fd == NULL)
+    return NULL;
+  state->size_T = 4;
+  state->T = malloc (state->size_T * sizeof (mpres_t));
+  if (state->T == NULL)
+    {
+      free (state->fd);
+      return NULL;;
+    }
+  state->X = X;
+  
+  for (i = 0; i < state->size_fd; i++)
+    {
+      MEMORY_TAG;
+      mpres_init (state->fd[i].x, modulus);
+      MEMORY_TAG;
+      mpres_init (state->fd[i].y, modulus);
+      MEMORY_UNTAG;
+    }
+
+  for (i = 0; i < state->size_T; i++)
+    {
+      MEMORY_TAG;
+      mpres_init (state->T[i], modulus);
+      MEMORY_UNTAG;
+    }
+
+  MPZ_INIT (t);
+  
+  mpres_add_ui (state->T[0], X->A, 2, modulus);
+  mpres_div_2exp (state->T[0], state->T[0], 2, modulus);
+  /* state->T[0] = b, ...T[1..3] = temp space for add3() */
+  
+  outputf (OUTPUT_TRACE, "ecm_rootsG_Mont_init: X = (%Zd::%Zd)\n",
+           X->x, X->y);
+  
+  mpres_set (state->fd[2].x, X->x, modulus);
+  mpres_set (state->fd[2].y, X->y, modulus);
+  mpz_set_ui (t, root_params->d1);
+  ecm_mul (state->fd[2].x, state->fd[2].y, t, modulus, state->T[0]);
+  /* fd[2] = d1*X */
+  outputf (OUTPUT_TRACE, "ecm_rootsG_Mont_init: d1*X = (%Zd::%Zd)\n",
+           state->fd[2].x, state->fd[2].y);
+  
+  mpres_set (state->fd[0].x, state->fd[2].x, modulus);
+  mpres_set (state->fd[0].y, state->fd[2].y, modulus);
+  ecm_mul (state->fd[0].x, state->fd[0].y, root_params->i0, modulus, 
+	   state->T[0]);
+  /* fd[0] = i0*d1*X */
+  outputf (OUTPUT_TRACE, "ecm_rootsG_Mont_init: i0*d1*X = (%Zd::%Zd)\n",
+           state->fd[0].x, state->fd[0].y);
+
+  mpres_set (state->fd[1].x, state->fd[2].x, modulus);
+  mpres_set (state->fd[1].y, state->fd[2].y, modulus);
+  mpz_sub_ui (t, root_params->i0, 1);
+  ecm_mul (state->fd[1].x, state->fd[1].y, t, modulus, state->T[0]);
+  /* fd[1] = (i0-1)*d1*X */
+  outputf (OUTPUT_TRACE, "ecm_rootsG_Mont_init: (i0-1)*d1*X = (%Zd::%Zd)\n",
+           state->fd[1].x, state->fd[1].y);
+  
+  st = elltime (st, cputime ());
+  outputf (OUTPUT_VERBOSE,
+	   "Initializing table of differences for G took %ldms\n", st);
+  
+  mpz_clear (t);
+
+  return state;
+}
+#endif
+
+ecm_roots_state *
+ecm_rootsG_init (mpz_t f, curve *X, root_params_t *root_params, 
+                 unsigned long dF, unsigned long blocks, mpmod_t modulus)
+{
+  unsigned int k, phid2;
   unsigned long muls = 0, gcds = 0;
   listz_t coeffs;
   ecm_roots_state *state;
@@ -598,64 +836,76 @@ ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
   int dickson_a;
   unsigned int T_inv;
   double bestnr;
-  unsigned int st = 0;
+  long st = 0;
 
-  ASSERT (gcd (d1, d2) == 1);
+  ASSERT (gcd (root_params->d1, root_params->d2) == 1);
+
+#ifdef MONT_ROOTS
+  if (root_params->S == 1)
+    return ecm_rootsG_Mont_init (X, root_params, dF, blocks, modulus);
+#endif
 
   if (test_verbose (OUTPUT_VERBOSE))
     st = cputime ();
   
+  state = (ecm_roots_state *) malloc (sizeof (ecm_roots_state));
+  if (state == NULL)
+    {
+      mpz_set_si (f, -1);
+      return NULL;
+    }
+
   /* If S < 0, use degree |S| Dickson poly, otherwise use x^S */
-  dickson_a = (S < 0) ? -1 : 0;
-  S = abs (S);
+  dickson_a = (root_params->S < 0) ? -1 : 0;
+  state->S = abs (root_params->S);
 
   /* Estimate the cost of a modular inversion (in unit of time per 
      modular multiplication) */
-  if (modulus->repr == MOD_BASE2)
+  if (modulus->repr == ECM_MOD_BASE2)
     T_inv = 18;
   else
     T_inv = 6;
   
   /* Guesstimate a value for the number of disjoint progressions to use */
   bestnr = -(4. + T_inv) + sqrt(12. * (double) dF * (double) blocks * 
-        (T_inv - 3.) * log (2. * d1) / log (2.) - (4. + T_inv) * (4. + T_inv));
-  bestnr /= 6. * (double) S * log (2. * d1) / log (2.0);
+        (T_inv - 3.) * log (2. * root_params->d1) / log (2.) - (4. + T_inv) * 
+        (4. + T_inv));
+  bestnr /= 6. * (double) (state->S) * log (2. * root_params->d1) / log (2.0);
   
   outputf (OUTPUT_TRACE, "ecm_rootsG_init: bestnr = %f\n", bestnr);
   
-  state = (ecm_roots_state *) malloc (sizeof (ecm_roots_state));
-  if (state == NULL)
-    goto exit_error;
-
   if (bestnr < 1.)
     state->nr = 1;
   else
     state->nr = (unsigned int) (bestnr + .5);
 
-  phid2 = phi (d2);
+  phid2 = eulerphi (root_params->d2);
 
-  /* Round up state->nr to multiple of phi(d2) */
+  /* Round up state->nr to multiple of eulerphi(d2) */
   if (phid2 > 1)
     state->nr = ((state->nr + (phid2 - 1)) / phid2) * phid2;
 
-  state->S = S;
   state->size_fd = state->nr * (state->S + 1);
 
-  outputf (OUTPUT_DEVVERBOSE, "ecm_rootsG_init: s=%Zd, d1=%u, d2=%d, dF=%d, blocks=%d, S=%u, T_inv = %d, nr=%d\n", 
-	   s, d1, d2, dF, blocks, S, T_inv, state->nr);
+  outputf (OUTPUT_DEVVERBOSE, "ecm_rootsG_init: i0=%Zd, d1=%lu, d2=%lu, "
+           "dF=%lu, blocks=%lu, S=%u, T_inv = %d, nr=%d\n", root_params->i0, 
+           root_params->d1, root_params->d2, dF, blocks, state->S, T_inv, 
+           state->nr);
   
   state->X = X;
   state->next = 0;
   state->dsieve = 1; /* We only init progressions coprime to d2, so nothing to be skipped */
-  state->rsieve = 1;
+  state->rsieve = 0;
+#ifdef MONT_ROOTS
+  state->form = EC_WEIERSTRASS_FORM;
+#endif
 
-  coeffs = init_progression_coeffs (s, d2, d1, state->nr / phid2, 1, S, 
-                                    dickson_a);
+  coeffs = init_progression_coeffs (root_params->i0, root_params->d2, 
+           root_params->d1, state->nr / phid2, 1, state->S, dickson_a);
 
   if (coeffs == NULL) /* error */
     {
       free (state);
-    exit_error:
       mpz_set_si (f, -1);
       return NULL;
     }
@@ -670,12 +920,15 @@ ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
     }
   for (k = 0; k < state->size_fd; k++)
     {
+      MEMORY_TAG;
       mpres_init (state->fd[k].x, modulus);
+      MEMORY_TAG;
       mpres_init (state->fd[k].y, modulus);
+      MEMORY_UNTAG;
     }
   
-  lenT = state->size_fd + 4;
-  state->T = (mpres_t *) malloc (lenT * sizeof (mpres_t));
+  state->size_T = state->size_fd + 4;
+  state->T = (mpres_t *) malloc (state->size_T * sizeof (mpres_t));
   if (state->T == NULL)
     {
       for (k = 0; k < state->size_fd; k++)
@@ -688,11 +941,15 @@ ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
       mpz_set_si (f, -1);
       return NULL;
     }
-  for (k = 0; k < lenT; k++)
-    mpres_init (state->T[k], modulus);
+  for (k = 0; k < state->size_T; k++)
+    {
+      MEMORY_TAG;
+      mpres_init (state->T[k], modulus);
+      MEMORY_UNTAG;
+    }
 
-  for (k = S + 1; k < state->size_fd; k += S + 1)
-     mpz_set_ui (coeffs[k + S], 1);
+  for (k = state->S + 1; k < state->size_fd; k += state->S + 1)
+     mpz_set_ui (coeffs[k + state->S], 1);
 
   if (test_verbose (OUTPUT_TRACE))
     for (k = 0; k < state->size_fd; k++)
@@ -704,10 +961,10 @@ ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
   if (youpi == ECM_ERROR)
     mpz_set_si (f, -1); /* fall through */
 
-  for (k = S + 1; k < state->size_fd; k += S + 1)
+  for (k = state->S + 1; k < state->size_fd; k += state->S + 1)
     {
-      mpres_set (state->fd[k + S].x, state->fd[S].x, modulus);
-      mpres_set (state->fd[k + S].y, state->fd[S].y, modulus);
+      mpres_set (state->fd[k + state->S].x, state->fd[state->S].x, modulus);
+      mpres_set (state->fd[k + state->S].y, state->fd[state->S].y, modulus);
     }
   
   clear_list (coeffs, state->size_fd);
@@ -718,7 +975,7 @@ ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
       if (youpi == ECM_FACTOR_FOUND_STEP2)
         outputf (OUTPUT_VERBOSE, "Found factor while computing fd[]\n");
 
-      ecm_rootsG_clear (state, S, modulus);
+      ecm_rootsG_clear (state, modulus);
       
       /* Signal that a factor was found, or an error occurred (f=-1) */
       state = NULL;
@@ -729,7 +986,7 @@ ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
         {
           st = elltime (st, cputime ());
           outputf (OUTPUT_VERBOSE,
-		   "Initializing table of differences for G took %ums", st);
+		   "Initializing table of differences for G took %ldms", st);
 	  outputf (OUTPUT_DEVVERBOSE, ", %lu muls and %lu extgcds",
 		   muls, gcds);
           outputf (OUTPUT_VERBOSE, "\n");
@@ -740,8 +997,7 @@ ecm_rootsG_init (mpz_t f, curve *X, mpz_t s, unsigned int d1, unsigned int d2,
 }
 
 void 
-ecm_rootsG_clear (ecm_roots_state *state, ATTRIBUTE_UNUSED int S, 
-                  ATTRIBUTE_UNUSED mpmod_t modulus)
+ecm_rootsG_clear (ecm_roots_state *state, ATTRIBUTE_UNUSED mpmod_t modulus)
 {
   unsigned int k;
   
@@ -752,7 +1008,7 @@ ecm_rootsG_clear (ecm_roots_state *state, ATTRIBUTE_UNUSED int S,
     }
   free (state->fd);
   
-  for (k = 0; k < state->size_fd + 4; k++)
+  for (k = 0; k < state->size_T; k++)
     mpres_clear (state->T[k], modulus);
   free (state->T);
   
@@ -773,15 +1029,26 @@ ecm_rootsG_clear (ecm_roots_state *state, ATTRIBUTE_UNUSED int S,
 */
 
 int 
-ecm_rootsG (mpz_t f, listz_t G, unsigned int dF, ecm_roots_state *state, 
+ecm_rootsG (mpz_t f, listz_t G, unsigned long dF, ecm_roots_state *state, 
             mpmod_t modulus)
 {
-  unsigned int i;
+  unsigned long i;
   unsigned long muls = 0, gcds = 0;
   int youpi = ECM_NO_FACTOR_FOUND;
-  unsigned int st;
+  long st;
+  point *fd = state->fd; /* to save typing */
   
   st = cputime ();
+
+  outputf (OUTPUT_TRACE, "ecm_rootsG: dF = %lu, state: nr = %u, next = %u, "
+           "S = %u, dsieve = %u, rsieve = %u,\n\tdickson_a = %d\n", 
+           dF, state->nr, state->next, state->S, state->dsieve, 
+           state->rsieve, state->dickson_a);
+#ifdef MONT_ROOTS
+  outputf (OUTPUT_TRACE, "form = %s, i0 = %ld\n", 
+           (state->form == EC_WEIERSTRASS_FORM) ? 
+           "Weierstrass" : "Montgomery", state->i0);
+#endif
   
   for (i = 0; i < dF;)
     {
@@ -789,10 +1056,60 @@ ecm_rootsG (mpz_t f, listz_t G, unsigned int dF, ecm_roots_state *state,
       if (state->next == state->nr)
         {
           /* Yes, time to update again */
-          youpi = addWnm (f, state->fd, state->X, modulus, state->nr, 
-                          state->S, state->T, &muls, &gcds);
-	  ASSERT(youpi != ECM_ERROR); /* no error can occur in addWnm */
-          state->next = 0;
+#ifdef MONT_ROOTS
+	  if (state->form == EC_MONTGOMERY_FORM) /* If in Montgomery form */
+	    {
+              mpres_swap (fd[0].x, fd[1].x, modulus);
+              mpres_swap (fd[0].y, fd[1].y, modulus);
+	      if (state->i0 == -1)
+	        {
+	          /* New point is 0*d1*X */
+	          mpres_set_ui (fd[0].x, 0, modulus);
+	          mpres_set_ui (fd[0].y, 0, modulus);
+	        }
+	      else if (state->i0 == 0)
+	        {
+	          /* Test that old point is (0::0) */
+	          ASSERT (mpres_is_zero (fd[1].x, modulus) && 
+	                  mpres_is_zero (fd[1].y, modulus));
+	          /* New point is 1*d1*X */
+	          mpres_set (fd[0].x, fd[2].x, modulus);
+	          mpres_set (fd[0].y, fd[2].y, modulus);
+	        }
+	      else if (state->i0 == 1)
+	        {
+	          /* New point is 2*d1*X */
+#if 0 && defined (WANT_ASSERT)
+                  /* Compare that fd[1] and fd[2] are identical */
+                  /* FIXME: strictly speaking, we need to normalize first! */
+                  mpres_sub (state->T[1], fd[1].x, fd[2].x, modulus);
+                  mpres_sub (state->T[2], fd[1].y, fd[2].y, modulus);
+                  ASSERT (mpres_is_zero (state->T[1], modulus) && 
+                          mpres_is_zero (state->T[2], modulus));
+#endif
+                  ASSERT (mpres_is_zero (fd[0].x, modulus));
+                  ASSERT (mpres_is_zero (fd[0].y, modulus));
+	          duplicate (fd[0].x, fd[0].y, fd[1].x, fd[1].y, modulus,
+	                     state->T[0], state->T[1], state->T[2], 
+	                     state->T[3]);
+	        }
+	      else 
+	        {
+	          add3 (fd[0].x, fd[0].y, fd[1].x, fd[1].y, fd[2].x, fd[2].y, 
+		        fd[0].x, fd[0].y, modulus, 
+		        state->T[1], state->T[2], state->T[3] );
+	        }
+	      if (state->i0 < 2)
+	        state->i0++;
+	    }
+	  else /* in Weierstrass form */
+#endif
+	    {
+	      youpi = addWnm (f, fd, state->X, modulus, state->nr, 
+			      state->S, state->T, &muls, &gcds);
+	      ASSERT(youpi != ECM_ERROR); /* no error can occur in addWnm */
+	    }
+	  state->next = 0;
           
           if (youpi == ECM_FACTOR_FOUND_STEP2)
             {
@@ -803,17 +1120,211 @@ ecm_rootsG (mpz_t f, listz_t G, unsigned int dF, ecm_roots_state *state,
       
       /* Is this a root we should skip? (Take only if gcd == 1) */
       if (gcd (state->rsieve, state->dsieve) == 1)
-        mpres_get_z (G[i++], (state->fd + state->next * (state->S + 1))->x, 
-                       modulus);
+	{
+#ifdef MONT_ROOTS
+	  if (state->form == EC_MONTGOMERY_FORM)
+	    {
+	      if (state->i0 != 0)
+	        {
+                  /* Normalize coordinates to z=1 */
+                  if (!mpres_invert (state->T[1], fd[0].y, modulus))
+                    {
+                      mpres_gcd (f, fd[0].y, modulus);
+	              outputf (OUTPUT_VERBOSE, 
+	                       "Found factor while computing G[%lu]\n", i);
+                      return ECM_FACTOR_FOUND_STEP2;
+                    }
+                  mpres_mul (state->T[1], state->T[1], 
+                             (fd + state->next * (state->S + 1))->x,
+                             modulus);
+                }
+              mpres_get_z (G[i++], state->T[1], modulus);
+            } 
+          else 
+#endif
+            {
+              mpres_get_z (G[i++], 
+                           (fd + state->next * (state->S + 1))->x, 
+                           modulus);
+            }
+	  outputf (OUTPUT_TRACE, 
+	           "ecm_rootsG: storing d1*%u*X = %Zd in G[%lu]\n",
+		   state->rsieve, G[i - 1], i);
+	}
 
       state->next ++;
       state->rsieve ++;
     }
   
-  outputf (OUTPUT_VERBOSE, "Computing roots of G took %ums",
+  outputf (OUTPUT_VERBOSE, "Computing roots of G took %ldms",
 	   elltime (st, cputime ()));
   outputf (OUTPUT_DEVVERBOSE, ", %lu muls and %lu extgcds", muls, gcds);
   outputf (OUTPUT_VERBOSE, "\n");
   
   return youpi;
 }
+
+
+/* Find smallest i >= 0 so that 
+   f(j * d2) X = f((i0 + i) * d1) X over GF(p) */
+long 
+ecm_findmatch (const unsigned long j, root_params_t *root_params, 
+               curve *X, mpmod_t n, mpz_t p)
+{
+  const int dickson_a = root_params->S < 0 ? -1 : 0;
+  const unsigned int S = abs (root_params->S);
+  const int sizeT = S + 3;
+  long i = 0, k;
+  point iX, jX;
+  curve Xp; /* The point and curve over GF(p) */
+  mpmod_t modulus;
+  mpz_t s, t; /* temp vars */
+  mpres_t u, v; /* temp vars */
+  listz_t coeffs;
+  point *fd;
+  mpres_t *T;
+  
+  outputf (OUTPUT_RESVERBOSE, "Looking for i so that "
+           "f(i*%lu)*X = f(%lu*%lu)*X\n", root_params->d1, 
+           j, root_params->d2);
+  
+  mpmod_init (modulus, p, 0);
+  mpz_init (s);
+  mpz_init (t);
+  mpres_init (u, modulus);
+  mpres_init (v, modulus);
+  mpres_init (Xp.x, modulus);
+  mpres_init (Xp.y, modulus);
+  mpres_init (Xp.A, modulus);
+  mpres_init (iX.x, modulus);
+  mpres_init (iX.y, modulus);
+  mpres_init (jX.x, modulus);
+  mpres_init (jX.y, modulus);
+  T = malloc (sizeT * sizeof (mpres_t));
+  if (T == NULL)
+    goto clear_and_exit;
+  for (k = 0; k < sizeT; k++)
+    mpres_init (T[k], modulus);
+  fd = malloc ((S + 1) * sizeof (point));
+  if (fd == NULL)
+    goto clear_T_and_exit;
+  for (k = 0; k < S + 1; k++)
+    {
+      mpres_init (fd[k].x, modulus);
+      mpres_init (fd[k].y, modulus);
+    }
+
+  /* Copy the parameters of the curve over Z/ZN to the curve over GF(p) */
+  mpres_get_z (t, X->x, n);
+  mpres_set_z (Xp.x, t, modulus);
+  mpres_get_z (t, X->y, n);
+  mpres_set_z (Xp.y, t, modulus);
+  mpres_get_z (t, X->A, n);
+  mpres_set_z (Xp.A, t, modulus);
+  
+  /* We use init_progression_coeffs() to compute f(j * d2) */
+  mpz_set_ui (t, j);
+  coeffs = init_progression_coeffs (t, 1UL, root_params->d2, 1U, 1U, S, 
+                                    dickson_a);
+  if (coeffs == NULL)
+    goto clear_fd_and_exit;
+  
+  /* Now compute f(j * d2) X */
+  multiplyW2n (NULL, &jX, &Xp, coeffs, 1U, modulus, u, v, T, NULL, NULL);
+
+  clear_list (coeffs, S + 1);
+
+  /* We'll keep {f(j * d2) X}_x in s */
+  mpres_get_z (s, jX.x, modulus);
+  outputf (OUTPUT_DEVVERBOSE, "ecm_findmatch: (f(j * d2) X)_x = %Zd\n", s);
+
+  /* Now compute {f((i0 + i) d1) X}_x one at a time and put them in t, 
+     until s == t */
+  
+  /* Init the progression */
+  coeffs = init_progression_coeffs (root_params->i0, 1UL, root_params->d1, 
+                                    1U, 1U, S, dickson_a);
+  if (coeffs == NULL)
+    goto clear_fd_and_exit;
+  multiplyW2n (NULL, fd, &Xp, coeffs, S + 1, modulus, u, v, T, NULL, NULL);
+  clear_list (coeffs, S + 1);
+  
+  i = 0;
+  mpres_get_z (t, fd[0].x, modulus);
+  while (mpz_cmp (s, t) != 0)
+    {
+      addWnm (NULL, fd, &Xp, modulus, 1, S, T, NULL, NULL);
+      mpres_get_z (t, fd[0].x, modulus);
+      i++;
+    }
+
+  outputf (OUTPUT_DEVVERBOSE, "ecm_findmatch: i - i0 = %ld, "
+           "{f(i * d1) X}_x = %Zd\n", i, t);
+
+  /* We'll compute f(i * d1) and compare it to f(j * d2) to verify 
+     correctness of the result, and to determine whether it was
+     f(i * d1)-f(j * d2) or f(i * d1)+f(j * d2) that found the factor */
+  /* We use init_progression_coeffs() to compute f(i * d1) */
+  mpz_add_ui (t, root_params->i0, i);
+  coeffs = init_progression_coeffs (t, 1UL, root_params->d1, 1U, 1U, S, 
+                                    dickson_a);
+  if (coeffs == NULL)
+    goto clear_fd_and_exit;
+  
+  /* Now compute iX = f(i * d1) X */
+  multiplyW2n (NULL, &iX, &Xp, coeffs, 1U, modulus, u, v, T, NULL, NULL);
+  
+  clear_list (coeffs, S + 1);
+
+  mpres_get_z (t, iX.x, modulus);
+  if (mpz_cmp (s, t) != 0)
+    {
+      outputf (OUTPUT_ERROR, "ecm_findmatch: ERROR, (f(i*d1) X)_x != "
+               "(f(j*d2) X)_x\n(f(i*d1) X)_x = %Zd\n", t);
+      i = 0;
+    }
+
+  mpres_get_z (s, jX.y, modulus);
+  mpres_get_z (t, iX.y, modulus);
+  if (mpz_cmp (s, t) != 0)
+    {
+      mpz_sub (t, p, t);
+      if (mpz_cmp (s, t) != 0)
+        {
+          mpz_sub (t, p, t);
+          outputf (OUTPUT_ERROR, "ecm_findmatch: ERROR, (f(i*d1) X)_y != "
+                   "+-(f(j*d2) X)_y\n");
+          outputf (OUTPUT_ERROR, "(f(i*d1) X)_y = %Zd\n", t);
+          outputf (OUTPUT_ERROR, "(f(j*d2) X)_y = %Zd\n", s);
+        }
+      else
+        i = -i;
+    }
+
+clear_fd_and_exit:
+  for (k = 0; k < S + 1; k++)
+    {
+      mpres_clear (fd[k].x, modulus);
+      mpres_clear (fd[k].y, modulus);
+    }
+  free(fd);
+clear_T_and_exit:
+  for (k = 0; k < sizeT; k++)
+    mpres_clear (T[k], modulus);
+  free (T);
+clear_and_exit:
+  mpz_clear (s);
+  mpz_clear (t);
+  mpres_clear (u, modulus);
+  mpres_clear (v, modulus);
+  mpres_clear (Xp.x, modulus);
+  mpres_clear (Xp.y, modulus);
+  mpres_clear (Xp.A, modulus);
+  mpres_clear (iX.x, modulus);
+  mpres_clear (iX.y, modulus);
+  mpres_clear (jX.x, modulus);
+  mpres_clear (jX.y, modulus);
+  mpmod_clear (modulus);
+  
+  return i;
+}
