@@ -21,16 +21,18 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#if !defined (_MSC_VER)
-#include <strings.h> /* for strncasecmp */
-#include <unistd.h>
-#endif
-#include <ctype.h>
 #include <time.h>
-#include "gmp.h"
-#include "ecm.h"
 #include "ecm-ecm.h"
 
+#if HAVE_STRINGS_H
+# include <strings.h> /* for strncasecmp */
+#endif
+
+#if HAVE_CTYPE_H
+# include <ctype.h>
+#endif
+
+
 /*****************************************************************
  *   Syntax for this VERY simple recursive expression parser:	 *
  *                                                               *
@@ -63,9 +65,11 @@ static void eval_sum (mpz_t prior_n, mpz_t n,char op);
 static int  eval_Phi (mpz_t prior_n, mpz_t n, int ParamCnt);
 static int  eval_2 (int bInFuncParams);
 
+#if 0 /* strncasecmp is a required function in configure.in */
 #if defined (_MSC_VER) || defined (__MINGW32__)
 #define strncasecmp strnicmp
 #endif
+#endif
 
 /**************************************/
 /* Main expression evalation function */
@@ -114,7 +118,7 @@ ChompLine:;
 
       /* strip space and tabs out here, and then we DON'T have to mess with them in the rest of the parser */
       if (!isspace (c))
-	expr[nCurSize++] = c;
+	expr[nCurSize++] = (char) c;
 
       if (nCurSize == nMaxSize)
       {
@@ -349,7 +353,7 @@ int eval_Phi (mpz_t b, mpz_t n, int ParamCnt)
 {
   int factors[200];
   unsigned dwFactors=0, dw;
-  int N, B;
+  int B;
   double p;
   mpz_t D, T, org_n;
 
@@ -391,7 +395,6 @@ int eval_Phi (mpz_t b, mpz_t n, int ParamCnt)
 
   /* Ok, do the real h_primative work, since we are not one of the trivial case */
 
-  N = mpz_get_si(n);
   B = mpz_get_si(b);
 
   if (mpz_cmp_ui(b, B))
