@@ -19,11 +19,10 @@ along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 MA 02111-1307, USA. */
 
-#ifdef MEMORY_DEBUG
-
+#include "config.h"
 #include <stdio.h>
 #include <stdlib.h>  /* for abort */
-#include "gmp.h"
+#include <gmp.h>
 #include "ecm.h"
 
 void *__gmp_default_allocate (size_t);
@@ -40,14 +39,23 @@ void __gmp_default_free (void *, size_t);
    when tests_memory_end() is called.  Test programs must be sure to have
    "clear"s for all temporary variables used.  */
 
+#define NAME_LEN 8
+
 struct header {
   void           *ptr;
   size_t         size;
+  char           name[NAME_LEN];
+  unsigned int   line;
   struct header  *next;
 };
 
 struct header  *tests_memory_list = NULL;
 
+static unsigned long nr_realloc = 0, nr_realloc_move = 0;
+static char cur_name[NAME_LEN];
+static unsigned int cur_line;
+static unsigned long cur_mem, peak_mem;
+
 /* Return a pointer to a pointer to the found block (so it can be updated
    when unlinking). */
 static struct header **
@@ -74,6 +82,7 @@ static void *
 tests_allocate (size_t size)
 {
   struct header  *h;
+  int i;
 
   if (size == 0)
     {
@@ -81,12 +90,21 @@ tests_allocate (size_t size)
       abort ();
     }
 
+  if (cur_name[0] == 0)
+    cur_name[1] = 0; /* Set breakpoint here to find untagged allocs */
+
   h = (struct header *) __gmp_default_allocate (sizeof (*h));
   h->next = tests_memory_list;
   tests_memory_list = h;
 
   h->size = size;
   h->ptr = (struct header*) __gmp_default_allocate (size);
+  for (i = 0; i < NAME_LEN; i++)
+    h->name[i] = cur_name[i];
+  h->line = cur_line;
+  cur_mem += size;
+  if (cur_mem > peak_mem)
+    peak_mem = cur_mem;
   return h->ptr;
 }
 
@@ -113,13 +131,33 @@ tests_reallocate (void *ptr, size_t old_size, size_t new_size)
 
   if (h->size != old_size)
     {
-      printf ("tests_reallocate(): bad old size %u, should be %u\n",
+      printf ("tests_reallocate(): bad old size %zd, should be %zd\n",
               old_size, h->size);
       abort ();
     }
 
+  if (h->size > cur_mem)
+    {
+      printf ("tests_reallocate(): h->size = %zd but cur_mem = %lu\n",
+              h->size, cur_mem);
+      abort();
+    }
+
+  cur_mem = cur_mem - h->size + new_size;
+  if (cur_mem > peak_mem)
+    peak_mem = cur_mem;
+
+#if 0
+  printf ("Reallocating %p, first allocated in %s, line %d, from %d to %d\n",
+          ptr, h->name, h->line, h->size, new_size);
+  if (new_size <= h->size)
+    printf ("Unnecessary realloc!\n");
+#endif
+  nr_realloc++;
   h->size = new_size;
   h->ptr = (struct header*) __gmp_default_reallocate (ptr, old_size, new_size);
+  if (h->ptr != ptr)
+    nr_realloc_move++;
   return h->ptr;
 }
 
@@ -142,6 +180,15 @@ tests_free_nosize (void *ptr)
   struct header  **hp = tests_free_find (ptr);
   struct header  *h = *hp;
 
+  if (h->size > cur_mem)
+    {
+      printf ("tests_free_nosize(): h->size = %zd but cur_mem = %lu\n",
+              h->size, cur_mem);
+      abort();
+    }
+
+  cur_mem -= h->size;
+
   *hp = h->next;  /* unlink */
 
   __gmp_default_free (ptr, h->size);
@@ -156,7 +203,7 @@ tests_free (void *ptr, size_t size)
 
   if (h->size != size)
     {
-      printf ("tests_free(): bad size %u, should be %u\n", size, h->size);
+      printf ("tests_free(): bad size %zd, should be %zd\n", size, h->size);
       abort ();
     }
 
@@ -167,6 +214,10 @@ void
 tests_memory_start (void)
 {
   mp_set_memory_functions (tests_allocate, tests_reallocate, tests_free);
+  cur_name[0] = 0;
+  cur_line = 0;
+  cur_mem = 0L;
+  peak_mem = 0L;
 }
 
 void
@@ -188,11 +239,25 @@ tests_memory_end (void)
 
       count = 0;
       for (h = tests_memory_list; h != NULL; h = h->next)
-	count++;
+        {
+	  count++;
+	  printf ("Memory at %p, allocated by %s, line %d\n", 
+	          h->ptr, h->name, h->line);
+	}
 
       printf ("    %u block(s) remaining\n", count);
       abort ();
     }
+  
+  if (cur_mem != 0)
+    {
+      printf ("tests_memory_end(): cur_mem = %lu but list of allocated "
+              "memory empty\n", cur_mem);
+      abort ();
+    }
+  
+  printf ("%lu reallocates, %lu reallocates with move, peak_mem = %lu\n", 
+          nr_realloc, nr_realloc_move, peak_mem);
 }
 
 void
@@ -212,19 +277,22 @@ tests_memory_status (void)
 
     }
 
+  if (size != cur_mem)
+    {
+      printf ("tests_memory_status(): size = %d but cur_mem = %lu", 
+              size, cur_mem);
+      abort();
+    }
+
   printf ("    %u blocks remaining, total size %u\n", count, size);
 }
 
-#else
-
-/* Stop compiler complaining about empty file */
-
-void memory_dummy (void);
-
-void 
-memory_dummy (void)
+void tests_memory_set_location (char *name, unsigned int line)
 {
-  return;
+  unsigned int i;
+
+  for (i = 0; i < NAME_LEN; i++)
+    cur_name[i] = name[i];
+  cur_line = line;
 }
 
-#endif
