@@ -20,40 +20,31 @@
   MA 02111-1307, USA.
 */
 
-#include <stdlib.h>
 #include <stdio.h>
-#include "gmp.h"
-
-/* For ECM_ERROR */
-#include "ecm.h"
-
+#include <stdlib.h>
 #include "ecm-gmp.h"
 #include "ecm-impl.h"
 
+#ifdef NATIVE_REDC
+  #include "asmredc.h"
+#endif
+
 FILE *ECM_STDOUT, *ECM_STDERR; /* define them here since needed in tune.c */
 
 /* define WANT_ASSERT to check normalization of residues */
 /* #define WANT_ASSERT 1 */
 /* #define DEBUG */
 
-#define ASSERT_NORMALIZED(x) ASSERT ((modulus->repr != MOD_MODMULN && \
-				      modulus->repr != MOD_REDC) || \
+#define ASSERT_NORMALIZED(x) ASSERT ((modulus->repr != ECM_MOD_MODMULN && \
+				      modulus->repr != ECM_MOD_REDC) || \
 			     mpz_size (x) <= mpz_size (modulus->orig_modulus))
 #define MPZ_NORMALIZED(x)    ASSERT (PTR(x)[ABSIZ(x)-1] != 0)
 
-#ifndef MPZMOD_THRESHOLD
-#define MPZMOD_THRESHOLD MPZMOD_THRESHOLD_DEFAULT
-#endif
-
-#ifndef REDC_THRESHOLD
-#define REDC_THRESHOLD REDC_THRESHOLD_DEFAULT
-#endif
-
-
 #ifndef GMP_NUMB_BITS
 #define GMP_NUMB_BITS __GMP_BITS_PER_MP_LIMB
 #endif
 
+
 void base2mod (mpres_t, mpres_t, mpres_t, mpmod_t);
 void base2mod_1 (mpres_t, mpres_t, mpmod_t);
 void REDC (mpres_t, mpres_t, mpz_t, mpmod_t);
@@ -70,15 +61,25 @@ isbase2 (mpz_t n, double threshold)
   int res = 0; 
   mpz_t u, w;
 
-  mpz_init (u);
-  mpz_init (w);
+  MPZ_INIT (u);
+  MPZ_INIT (w);
   lo = mpz_sizeinbase (n, 2) - 1; /* 2^lo <= n < 2^(lo+1) */  
   mpz_set_ui (u, 1);
   mpz_mul_2exp (u, u, 2 * lo);
   mpz_mod (w, u, n); /* 2^(2lo) mod n = -/+2^(2lo-l) if m*n = 2^l+/-1 */
-  if (mpz_cmp_ui (w, 1) == 0) /* if 2^(2lo) mod n = 1, then n divides 2^lo+1, 
-				 since n has lo+1 bits. */
-    return lo;
+  if (mpz_cmp_ui (w, 1) == 0) /* if 2^(2lo) mod n = 1, then n divides
+                                 2^(2lo)-1. If algebraic factors have been
+                                 removed, n divides either 2^lo+1 or 2^lo-1.
+                                 But since n has lo+1 bits, n can only divide
+                                 2^lo+1. More precisely, n must be 2^lo+1. */
+    {
+      /* check that n equals 2^lo+1. Since n divides 2^(2lo)-1, n is odd. */
+      if (mpz_scan1 (n, 1) != lo)
+        lo = 0;
+      mpz_clear (w);
+      mpz_clear (u);
+      return lo;
+    }
   k = mpz_sizeinbase (w, 2) - 1;
   /* if w = 2^k then n divides 2^(2*lo-k)-1 */
   mpz_set_ui (u, 1);
@@ -100,6 +101,9 @@ isbase2 (mpz_t n, double threshold)
   if (abs (res) > (int) (threshold * (double) lo)) 
     res = 0;
 
+  if (abs (res) < 16)
+    res = 0;
+
   return res;
 }
 
@@ -147,6 +151,7 @@ base2mod_1 (mpres_t RS, mpres_t t, mpmod_t modulus)
     }
 }
 
+#ifdef HAVE___GMPN_MUL_FFT
 /* Fermat-mod */
 static void
 base2mod_2 (mpres_t RS, mp_size_t n, mpz_t modulus)
@@ -163,12 +168,13 @@ base2mod_2 (mpres_t RS, mp_size_t n, mpz_t modulus)
           if ((rp[n] = mpn_sub_1 (rp, rp, n, rp[n])))
             rp[n] = mpn_add_1 (rp, rp, n, rp[n]);
           MPN_NORMALIZE(rp, s);
-          SIZ(RS) = (SIZ(RS) > 0) ? s : -s;
+          SIZ(RS) = (SIZ(RS) > 0) ? (int) s : (int) -s;
         }
       else /* should happen rarely */
         mpz_mod (RS, RS, modulus);
     }
 }
+#endif
 
 /* subquadratic REDC, at mpn level.
    {orig,n} is the original modulus.
@@ -190,7 +196,7 @@ ecm_redc_n (mp_ptr rp, mp_srcptr xp, mp_srcptr orig, mp_srcptr aux, mp_size_t n)
   mpn_mul_n (tp, up, orig, n);
   /* add {x, 2n} and {tp, 2n}. We know that {tp, n} + {xp, n} will give
      either 0, or a carry out. If xp[n-1] <> 0, then there is a carry. */
-#ifdef HAVE_NATIVE_mpn_add_nc
+#ifdef HAVE___GMPN_ADD_NC
   cy = __gmpn_add_nc (rp, tp + n, xp + n, n, (mp_limb_t) ((xp[n - 1]) ? 1 : 0));
 #else
   cy = mpn_add_n (rp, tp + n, xp + n, n);
@@ -213,13 +219,12 @@ REDC (mpres_t r, mpres_t x, mpz_t t, mpmod_t modulus)
   if (ABSIZ(x) == 2 * n)
     {
       mp_ptr rp;
-      if (ALLOC(r) < n)
-	_mpz_realloc (r, n);
+      MPZ_REALLOC (r, n);
       rp = PTR(r);
       ecm_redc_n (rp, PTR(x), PTR(modulus->orig_modulus), 
 		PTR(modulus->aux_modulus), n);
       MPN_NORMALIZE(rp, n);
-      SIZ(r) = (SIZ(x) > 0) ? n : -n;
+      SIZ(r) = (SIZ(x) > 0) ? (int) n : (int) -n;
       MPZ_NORMALIZED (r);
     }
   else
@@ -257,7 +262,6 @@ mod_div2exp (mpz_t c, unsigned int k, mpmod_t modulus)
   mpz_mul (modulus->temp1, modulus->temp2, c);
   mpz_mod (c, modulus->temp1, modulus->orig_modulus);
 }
-
 /* r <- c/R^nn mod n, where n has nn limbs, and R=2^GMP_NUMB_BITS.
    n must be odd.
    c must have space for at least 2*nn limbs.
@@ -299,9 +303,129 @@ ecm_redc_basecase (mpz_ptr r, mpz_ptr c, mpmod_t modulus)
   if (cy != 0)
     mpn_sub_n (rp, rp, np, nn); /* a borrow should always occur here */
   MPN_NORMALIZE (rp, nn);
-  SIZ(r) = SIZ(c) < 0 ? -nn : nn;
+  SIZ(r) = SIZ(c) < 0 ? (int) -nn : (int) nn;
 }
 
+#ifdef NATIVE_REDC
+static mp_limb_t
+mulredc(mp_limb_t *z, const mp_limb_t *x, const mp_limb_t *y,
+    const mp_limb_t *m, mp_size_t N, mp_limb_t invm, mp_limb_t *tmp)
+{
+  mp_limb_t cy;
+
+  switch (N) {
+   case 1:
+    cy = mulredc1(z, x[0], y[0], m[0], invm);
+    break;
+   case 2:
+    cy = mulredc2(z, x, y, m, invm);
+    break;
+   case 3:
+    cy = mulredc3(z, x, y, m, invm);
+    break;
+   case 4:
+    cy = mulredc4(z, x, y, m, invm);
+    break;
+   case 5: 
+    cy = mulredc5(z, x, y, m, invm);
+    break;
+   case 6: 
+    cy = mulredc6(z, x, y, m, invm);
+    break;
+   case 7: 
+    cy = mulredc7(z, x, y, m, invm);
+    break;
+   case 8:
+    cy = mulredc8(z, x, y, m, invm);
+    break;
+   case 9:
+    cy = mulredc9(z, x, y, m, invm);
+    break;
+   case 10:
+    cy = mulredc10(z, x, y, m, invm);
+    break;
+   case 11:
+    cy = mulredc11(z, x, y, m, invm);
+    break;
+   case 12:
+    cy = mulredc12(z, x, y, m, invm);
+    break;
+   case 13:
+    cy = mulredc13(z, x, y, m, invm);
+    break;
+   case 14:
+    cy = mulredc14(z, x, y, m, invm);
+    break;
+   case 15:
+    cy = mulredc15(z, x, y, m, invm);
+    break;
+   case 16:
+    cy = mulredc16(z, x, y, m, invm);
+    break;
+   case 17:
+    cy = mulredc17(z, x, y, m, invm);
+    break;
+   case 18:
+    cy = mulredc18(z, x, y, m, invm);
+    break;
+   case 19:
+    cy = mulredc19(z, x, y, m, invm);
+    break;
+   case 20:
+    cy = mulredc20(z, x, y, m, invm);
+    break;
+   default:
+    {
+      mpn_mul_n(tmp, x, y, N);
+      ecm_redc3(tmp, m, N, invm);
+      cy = mpn_add_n (z, tmp + N, tmp, N);
+    }
+  }
+  return cy;
+}
+#endif
+
+
+/* 
+ * Same as previous, but combined with mul (if in asm)
+ */
+static void 
+ecm_mulredc_basecase (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
+{
+#ifndef NATIVE_REDC
+  mpz_mul (modulus->temp1, S1, S2);
+  ecm_redc_basecase(R, modulus->temp1, modulus);
+#else
+  mp_ptr rp;
+  mp_ptr s1p, s2p;
+  mp_srcptr np;
+  mp_limb_t cy;
+  mp_size_t j, nn = modulus->bits / __GMP_BITS_PER_MP_LIMB;
+
+  ASSERT(ALLOC(R) >= nn);
+  ASSERT(ALLOC(S1) >= nn);
+  ASSERT(ALLOC(S2) >= nn);
+  rp = PTR(R);
+  s1p = PTR(S1);
+  s2p = PTR(S2);
+  np = PTR(modulus->orig_modulus);
+  for (j = ABSIZ(S1); j < nn; j++) 
+    s1p[j] = 0;
+  for (j = ABSIZ(S2); j < nn; j++) 
+    s2p[j] = 0;
+
+  cy = mulredc(rp, s1p, s2p, np, nn, modulus->Nprim, PTR(modulus->temp1));
+
+  /* the result of Montgomery's REDC is less than 2^Nbits + N,
+     thus at most one correction is enough */
+  if (cy != 0)
+    mpn_sub_n (rp, rp, np, nn); /* a borrow should always occur here */
+  MPN_NORMALIZE (rp, nn);
+  SIZ(R) = (SIZ(S1)*SIZ(S2)) < 0 ? (int) -nn : (int) nn;
+#endif
+}
+
+
 /* don't use base2 if repr == -1, i.e. -nobase2 */
 void 
 mpmod_init (mpmod_t modulus, mpz_t N, int repr)
@@ -322,7 +446,7 @@ mpmod_init (mpmod_t modulus, mpz_t N, int repr)
       outputf (OUTPUT_VERBOSE, "Using MODMULN\n");
       mpmod_init_MODMULN (modulus, N);
     }
-  else if (mpz_sizeinbase (N, 2) < REDC_THRESHOLD)
+  else if (mpz_size (N) < REDC_THRESHOLD)
     {
       outputf (OUTPUT_VERBOSE, "Using mpz_mod\n");
       mpmod_init_MPZ (modulus, N);
@@ -342,12 +466,12 @@ mpmod_init_MPZ (mpmod_t modulus, mpz_t N)
   int Nbits;
   
   mpz_init_set (modulus->orig_modulus, N);
-  modulus->repr = MOD_PLAIN;
+  modulus->repr = ECM_MOD_MPZ;
   
   Nbits = mpz_size (N) * __GMP_BITS_PER_MP_LIMB; /* Number of bits, rounded
                                                     up to full limb */
-  mpz_init2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
-  mpz_init2 (modulus->temp2, Nbits);
+  MPZ_INIT2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
+  MPZ_INIT2 (modulus->temp2, Nbits);
   
   return;
 }
@@ -358,13 +482,13 @@ mpmod_init_BASE2 (mpmod_t modulus, int base2, mpz_t N)
   int Nbits;
   
   mpz_init_set (modulus->orig_modulus, N);
-  modulus->repr = MOD_BASE2;
+  modulus->repr = ECM_MOD_BASE2;
   modulus->bits = base2;
 
   Nbits = mpz_size (N) * __GMP_BITS_PER_MP_LIMB; /* Number of bits, rounded
                                                     up to full limb */
-  mpz_init2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
-  mpz_init2 (modulus->temp2, Nbits);
+  MPZ_INIT2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
+  MPZ_INIT2 (modulus->temp2, Nbits);
   
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, abs (base2));
@@ -390,7 +514,7 @@ mpmod_init_BASE2 (mpmod_t modulus, int base2, mpz_t N)
       if (i == 1)
         {
           modulus->Fermat = base2;
-#ifdef HAVE_GWNUM
+#if defined(HAVE_GWNUM) && !defined (TUNE)
           if (modulus->Fermat >= GWTHRESHOLD)
             Fgwinit (modulus->Fermat);
 #endif
@@ -405,15 +529,17 @@ mpmod_init_MODMULN (mpmod_t modulus, mpz_t N)
 {
   int Nbits;
 
+  MEMORY_TAG;
   mpz_init_set (modulus->orig_modulus, N);
+  MEMORY_UNTAG;
   
-  modulus->repr = MOD_MODMULN;
+  modulus->repr = ECM_MOD_MODMULN;
   Nbits = mpz_size (N) * __GMP_BITS_PER_MP_LIMB; /* Number of bits, rounded
                                                     up to full limb */
   modulus->bits = Nbits;
 
-  mpz_init2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
-  mpz_init2 (modulus->temp2, Nbits);
+  MPZ_INIT2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
+  MPZ_INIT2 (modulus->temp2, Nbits);
 
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, __GMP_BITS_PER_MP_LIMB);
@@ -425,24 +551,24 @@ mpmod_init_MODMULN (mpmod_t modulus, mpz_t N)
   modulus->Nprim = mpz_getlimbn (modulus->temp2, 0);
     /* Now Nprim = -1/n (mod 2^bits_per_limb) */
 
-  mpz_init (modulus->R2);
+  MPZ_INIT (modulus->R2);
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, 2 * Nbits);
   mpz_mod (modulus->R2, modulus->temp1, modulus->orig_modulus);
   /* Now R2 = (2^bits)^2 (mod N) */
   
-  mpz_init (modulus->R3);
+  MPZ_INIT (modulus->R3);
   mpz_mul_2exp (modulus->temp1, modulus->R2, Nbits);
   mpz_mod (modulus->R3, modulus->temp1, modulus->orig_modulus);
   /* Now R3 = (2^bits)^3 (mod N) */
 
-  mpz_init (modulus->multiple);
+  MPZ_INIT (modulus->multiple);
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, Nbits);
   /* compute ceil(2^bits / N) */
   mpz_cdiv_q (modulus->temp1, modulus->temp1, modulus->orig_modulus);
   mpz_mul (modulus->multiple, modulus->temp1, modulus->orig_modulus);
-  /* Now multiple is the largest multiple of N >= 2^bits */
+  /* Now multiple is the smallest multiple of N >= 2^bits */
 }
 
 void 
@@ -454,14 +580,14 @@ mpmod_init_REDC (mpmod_t modulus, mpz_t N)
   mpz_init_set (modulus->orig_modulus, N);
   
   n = mpz_size (N);
-  modulus->repr = MOD_REDC;
+  modulus->repr = ECM_MOD_REDC;
   Nbits = n * __GMP_BITS_PER_MP_LIMB; /* Number of bits, rounded
                                                     up to full limb */
   modulus->bits = Nbits;
   
   mpz_init2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
   mpz_init2 (modulus->temp2, Nbits);
-  mpz_init (modulus->aux_modulus);
+  MPZ_INIT (modulus->aux_modulus);
 
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, Nbits);
@@ -473,23 +599,24 @@ mpmod_init_REDC (mpmod_t modulus, mpz_t N)
   /* ensure aux_modulus has n allocated limbs, for ecm_redc_n */
   if (ABSIZ(modulus->aux_modulus) < n)
     {
+      /* WARNING: _mpz_realloc does not keep the value!!! */
       _mpz_realloc (modulus->aux_modulus, n);
       MPN_ZERO (PTR(modulus->aux_modulus) + ABSIZ(modulus->aux_modulus),
 		n - ABSIZ(modulus->aux_modulus));
     }
 
-  mpz_init (modulus->R2);
+  MPZ_INIT (modulus->R2);
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, 2 * Nbits);
   mpz_mod (modulus->R2, modulus->temp1, modulus->orig_modulus);
   /* Now R2 = (2^bits)^2 (mod N) */
   
-  mpz_init (modulus->R3);
+  MPZ_INIT (modulus->R3);
   mpz_mul_2exp (modulus->temp1, modulus->R2, Nbits);
   mpz_mod (modulus->R3, modulus->temp1, modulus->orig_modulus);
   /* Now R3 = (2^bits)^3 (mod N) */
   
-  mpz_init (modulus->multiple);
+  MPZ_INIT (modulus->multiple);
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, Nbits);
   /* compute ceil(2^bits / N) */
@@ -504,15 +631,15 @@ mpmod_clear (mpmod_t modulus)
   mpz_clear (modulus->orig_modulus);
   mpz_clear (modulus->temp1);
   mpz_clear (modulus->temp2);
-  if (modulus->repr == MOD_MODMULN || modulus->repr == MOD_REDC)
+  if (modulus->repr == ECM_MOD_MODMULN || modulus->repr == ECM_MOD_REDC)
     {
       mpz_clear (modulus->R2);
       mpz_clear (modulus->R3);
       mpz_clear (modulus->multiple);
-      if (modulus->repr == MOD_REDC)
+      if (modulus->repr == ECM_MOD_REDC)
         mpz_clear (modulus->aux_modulus);
     }
-#ifdef HAVE_GWNUM
+#if defined(HAVE_GWNUM) && !defined (TUNE)
   if (modulus->Fermat >= GWTHRESHOLD)
     Fgwclear ();
 #endif
@@ -520,34 +647,53 @@ mpmod_clear (mpmod_t modulus)
   return;
 }
 
+void 
+mpmod_pausegw (ATTRIBUTE_UNUSED mpmod_t modulus)
+{
+#if defined(HAVE_GWNUM) && !defined (TUNE)
+  if (modulus->Fermat >= GWTHRESHOLD)
+    Fgwclear ();
+#endif
+}
+
+void 
+mpmod_contgw (ATTRIBUTE_UNUSED mpmod_t modulus)
+{
+#if defined(HAVE_GWNUM) && !defined (TUNE)
+  if (modulus->Fermat >= GWTHRESHOLD)
+    Fgwinit (modulus->Fermat);
+#endif
+}
+
 void 
 mpres_init (mpres_t R, mpmod_t modulus)
 {
   /* use mpz_sizeinbase since modulus->bits may not be initialized yet */
-  mpz_init2 (R, mpz_sizeinbase (modulus->orig_modulus, 2));
+  mpz_init2 (R, mpz_sizeinbase (modulus->orig_modulus, 2) + GMP_NUMB_BITS);
 }
 
 /* realloc R so that it has at least the same number of limbs as modulus */
 void
 mpres_realloc (mpres_t R, mpmod_t modulus)
 {
-  if (modulus->repr == MOD_MODMULN)
-    _mpz_realloc (R, modulus->bits / GMP_NUMB_BITS);
+  if (modulus->repr == ECM_MOD_MODMULN)
+    MPZ_REALLOC (R, modulus->bits / GMP_NUMB_BITS);
 }
 
 /* R <- BASE^EXP mod modulus.
    Assume EXP >= 0.
  */
 void 
-mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
+mpres_pow (mpres_t R, mpres_t BASE, mpz_t EXP, mpmod_t modulus)
 {
   ASSERT_NORMALIZED (BASE);
-  if (modulus->repr == MOD_PLAIN)
+  ASSERT (mpz_sgn (EXP) >= 0);
+  if (modulus->repr == ECM_MOD_MPZ)
     {
       mpz_powm (R, BASE, EXP, modulus->orig_modulus);
     }
-  else if (modulus->repr == MOD_BASE2 || modulus->repr == MOD_MODMULN ||
-           modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_BASE2 || modulus->repr == ECM_MOD_MODMULN ||
+           modulus->repr == ECM_MOD_REDC)
     {
       unsigned int expidx;
       mp_limb_t bitmask, expbits;
@@ -593,9 +739,9 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
             {
               mpz_mul (modulus->temp1, modulus->temp2, modulus->temp2); /* r = r^2 */
 
-              if (modulus->repr == MOD_BASE2)
+              if (modulus->repr == ECM_MOD_BASE2)
                 base2mod (modulus->temp2 , modulus->temp1, modulus->temp1, modulus);
-              else if (modulus->repr == MOD_MODMULN)
+              else if (modulus->repr == ECM_MOD_MODMULN)
                 {
                   ecm_redc_basecase (modulus->temp2, modulus->temp1, modulus);
                 }
@@ -605,9 +751,9 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
               if (expbits & bitmask)
                 { 
                   mpz_mul (modulus->temp1, modulus->temp2, BASE);
-                  if (modulus->repr == MOD_BASE2)
+                  if (modulus->repr == ECM_MOD_BASE2)
                     base2mod (modulus->temp2, modulus->temp1, modulus->temp1, modulus);
-                  else if (modulus->repr == MOD_MODMULN)
+                  else if (modulus->repr == ECM_MOD_MODMULN)
                     {
                       ecm_redc_basecase (modulus->temp2, modulus->temp1, modulus);
                     }
@@ -624,7 +770,7 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
       mpz_set (R, modulus->temp2); /* TODO: isn't it possible to use R instead
 				      of modulus->temp2 above to avoid this
 				      copy? */
-    } /* if (modulus->repr == MOD_BASE2 || ... ) */
+    } /* if (modulus->repr == ECM_MOD_BASE2 || ... ) */
   ASSERT_NORMALIZED (R);
 }
 
@@ -644,13 +790,13 @@ mpres_is_zero (mpres_t S, mpmod_t modulus)
 void 
 mpres_ui_pow (mpres_t R, unsigned int BASE, mpres_t EXP, mpmod_t modulus)
 {
-  if (modulus->repr == MOD_PLAIN)
+  if (modulus->repr == ECM_MOD_MPZ)
     {
       mpz_set_ui (modulus->temp1, BASE);
       mpz_powm (R, modulus->temp1, EXP, modulus->orig_modulus);
     }
-  else if (modulus->repr == MOD_BASE2 || modulus->repr == MOD_MODMULN ||
-           modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_BASE2 || modulus->repr == ECM_MOD_MODMULN ||
+           modulus->repr == ECM_MOD_REDC)
     {
       unsigned int expidx;
       mp_limb_t bitmask, expbits;
@@ -680,7 +826,7 @@ mpres_ui_pow (mpres_t R, unsigned int BASE, mpres_t EXP, mpmod_t modulus)
     /* bitmask is set to mask in the msb of expbits */
     
       mpz_set_ui (modulus->temp2, BASE); /* temp2 = BASE */
-      if (modulus->repr == MOD_MODMULN || modulus->repr == MOD_REDC)
+      if (modulus->repr == ECM_MOD_MODMULN || modulus->repr == ECM_MOD_REDC)
         {
           mpz_mul_2exp (modulus->temp1, modulus->temp2, modulus->bits);
           mpz_mod (modulus->temp2, modulus->temp1, modulus->orig_modulus);
@@ -693,9 +839,9 @@ mpres_ui_pow (mpres_t R, unsigned int BASE, mpres_t EXP, mpmod_t modulus)
             {
               mpz_mul (modulus->temp1, modulus->temp2, modulus->temp2); /* r = r^2 */
 
-              if (modulus->repr == MOD_BASE2)
+              if (modulus->repr == ECM_MOD_BASE2)
                 base2mod (modulus->temp2 , modulus->temp1, modulus->temp1, modulus);
-              else if (modulus->repr == MOD_MODMULN)
+              else if (modulus->repr == ECM_MOD_MODMULN)
                 {
                   ecm_redc_basecase (modulus->temp2, modulus->temp1, modulus);
                 }
@@ -725,7 +871,7 @@ mpres_ui_pow (mpres_t R, unsigned int BASE, mpres_t EXP, mpmod_t modulus)
         }
       mpz_set (R, modulus->temp2); /* TODO: use R instead of modulus->temp2
 				      above to avoid this copy? */
-    } /* if (modulus->repr == MOD_BASE2 || ... ) */
+    } /* if (modulus->repr == ECM_MOD_BASE2 || ... ) */
   ASSERT_NORMALIZED (R);
 }
 
@@ -735,23 +881,41 @@ mpres_mul (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
   ASSERT_NORMALIZED (S1);
   ASSERT_NORMALIZED (S2);
 
-#ifdef HAVE_GWNUM
-  if (modulus->repr == MOD_BASE2 && modulus->Fermat >= 1024)
+#if defined(HAVE_GWNUM) && !defined (TUNE)
+  if (modulus->repr == ECM_MOD_BASE2 && modulus->Fermat >= GWTHRESHOLD)
     {
       base2mod_1 (S1, modulus->temp1, modulus);
       base2mod_1 (S2, modulus->temp1, modulus);
+
+#ifdef DEBUG
+      mpz_mul (modulus->temp1, S1, S2);
+      base2mod_1 (modulus->temp1, modulus->temp2, modulus);
+      mpz_mod (modulus->temp2, modulus->temp1, modulus->orig_modulus);
+#endif
+
       ASSERT (mpz_sizeinbase (S1, 2) <= (unsigned) abs(modulus->bits));
       ASSERT (mpz_sizeinbase (S2, 2) <= (unsigned) abs(modulus->bits));
       Fgwmul (R, S1, S2);
+
+#ifdef DEBUG
+      mpz_mod (modulus->temp1, R, modulus->orig_modulus);
+      if (mpz_cmp (modulus->temp1, modulus->temp2) != 0)
+        {
+          fprintf (stderr, "mpres_mul: results of gwmul and mpz_mul differ\n");
+          gmp_fprintf (stderr, "GMP result   : %Zd\nGWNUM result : %Zd\n", 
+                       modulus->temp2, modulus->temp1);
+        }
+#endif
+
       return;
     }
-#elif defined(HAVE_FFT)
-  if (modulus->repr == MOD_BASE2 && modulus->Fermat >= 32768)
+#elif defined(HAVE___GMPN_MUL_FFT)
+  if (modulus->repr == ECM_MOD_BASE2 && modulus->Fermat >= 32768)
     {
       mp_size_t n = modulus->Fermat / __GMP_BITS_PER_MP_LIMB;
       unsigned long k;
       
-      _mpz_realloc (R, n + 1);
+      MPZ_REALLOC (R, n + 1);
       k = mpn_fft_best_k (n, S1 == S2);
       ASSERT(mpn_fft_next_size (n, k) == n);
       base2mod_2 (S1, n, modulus->orig_modulus);
@@ -759,22 +923,23 @@ mpres_mul (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
       mpn_mul_fft (PTR(R), n, PTR(S1), ABSIZ(S1), PTR(S2), ABSIZ(S2), k);
       n ++;
       MPN_NORMALIZE(PTR(R), n);
-      SIZ(R) = ((SIZ(S1) ^ SIZ(S2)) >= 0) ? n : -n;
+      SIZ(R) = ((SIZ(S1) ^ SIZ(S2)) >= 0) ? (int) n : (int) -n;
       return;
     }
 #endif
 
-  mpz_mul (modulus->temp1, S1, S2);
+  if (modulus->repr != ECM_MOD_MODMULN)
+    mpz_mul (modulus->temp1, S1, S2);
 
   switch (modulus->repr)
     {
-    case MOD_BASE2:
+    case ECM_MOD_BASE2:
       base2mod (R, modulus->temp1, modulus->temp1, modulus);
       break;
-    case MOD_MODMULN:
-      ecm_redc_basecase (R, modulus->temp1, modulus);
+    case ECM_MOD_MODMULN:
+      ecm_mulredc_basecase (R, S1, S2, modulus);
       break;
-    case MOD_REDC:
+    case ECM_MOD_REDC:
       REDC (R, modulus->temp1, modulus->temp2, modulus);
       break;
     default:
@@ -830,13 +995,13 @@ void
 mpres_add_ui (mpres_t R, mpres_t S, unsigned int n, mpmod_t modulus)
 {
   ASSERT_NORMALIZED (S);
-  if (modulus->repr == MOD_PLAIN || modulus->repr == MOD_BASE2)
+  if (modulus->repr == ECM_MOD_MPZ || modulus->repr == ECM_MOD_BASE2)
     {
       mpz_add_ui (R, S, n);
       if (mpz_cmp (R, modulus->orig_modulus) > 0)
         mpz_sub (R, R, modulus->orig_modulus);
     }
-  else if (modulus->repr == MOD_MODMULN || modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_MODMULN || modulus->repr == ECM_MOD_REDC)
     {
       mpz_set_ui (modulus->temp1, n);
       mpz_mul_2exp (modulus->temp1, modulus->temp1, modulus->bits);
@@ -853,7 +1018,7 @@ mpres_add (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
   ASSERT_NORMALIZED (S1);
   ASSERT_NORMALIZED (S2);
   mpz_add (R, S1, S2);
-  if ((modulus->repr == MOD_MODMULN || modulus->repr == MOD_REDC) &&
+  if ((modulus->repr == ECM_MOD_MODMULN || modulus->repr == ECM_MOD_REDC) &&
       ABSIZ(R) > ABSIZ(modulus->orig_modulus))
     {
       if (SIZ(R) > 0)
@@ -869,13 +1034,13 @@ void
 mpres_sub_ui (mpres_t R, mpres_t S, unsigned int n, mpmod_t modulus)
 {
   ASSERT_NORMALIZED (S);
-  if (modulus->repr == MOD_PLAIN || modulus->repr == MOD_BASE2)
+  if (modulus->repr == ECM_MOD_MPZ || modulus->repr == ECM_MOD_BASE2)
     {
       mpz_sub_ui (R, S, n);
       if (mpz_sgn (R) < 0)
         mpz_add (R, R, modulus->orig_modulus);
     }
-  else if (modulus->repr == MOD_MODMULN || modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_MODMULN || modulus->repr == ECM_MOD_REDC)
     {
       mpz_set_ui (modulus->temp1, n);
       mpz_mul_2exp (modulus->temp1, modulus->temp1, modulus->bits);
@@ -892,7 +1057,7 @@ mpres_sub (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
   ASSERT_NORMALIZED (S1);
   ASSERT_NORMALIZED (S2);
   mpz_sub (R, S1, S2);
-  if ((modulus->repr == MOD_MODMULN || modulus->repr == MOD_REDC) &&
+  if ((modulus->repr == ECM_MOD_MODMULN || modulus->repr == ECM_MOD_REDC) &&
       ABSIZ(R) > ABSIZ(modulus->orig_modulus))
     {
       if (SIZ(R) > 0)
@@ -907,17 +1072,17 @@ mpres_sub (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
 void 
 mpres_set_z (mpres_t R, mpz_t S, mpmod_t modulus)
 {
-  if (modulus->repr == MOD_PLAIN || modulus->repr == MOD_BASE2)
+  if (modulus->repr == ECM_MOD_MPZ || modulus->repr == ECM_MOD_BASE2)
     {
       mpz_mod (R, S, modulus->orig_modulus);
     }
-  else if (modulus->repr == MOD_MODMULN)
+  else if (modulus->repr == ECM_MOD_MODMULN)
     {
       mpz_mod (modulus->temp2, S, modulus->orig_modulus);
       mpz_mul (modulus->temp1, modulus->temp2, modulus->R2);
       ecm_redc_basecase (R, modulus->temp1, modulus);
     }
-  else if (modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_REDC)
     {
       mpz_mod (modulus->temp2, S, modulus->orig_modulus);
       mpz_mul (modulus->temp1, modulus->temp2, modulus->R2);
@@ -931,17 +1096,18 @@ void
 mpres_get_z (mpz_t R, mpres_t S, mpmod_t modulus)
 {
   ASSERT_NORMALIZED (S);
-  if (modulus->repr == MOD_PLAIN || modulus->repr == MOD_BASE2)
+  if (modulus->repr == ECM_MOD_MPZ || modulus->repr == ECM_MOD_BASE2)
     {
       mpz_mod (R, S, modulus->orig_modulus);
     }
-  else if (modulus->repr == MOD_MODMULN)
+  else if (modulus->repr == ECM_MOD_MODMULN)
     {
       mpz_set (modulus->temp1, S);
-      _mpz_realloc (R, modulus->bits / GMP_NUMB_BITS);
+      MPZ_REALLOC (R, modulus->bits / GMP_NUMB_BITS);
       ecm_redc_basecase (R, modulus->temp1, modulus);
+      mpz_mod (R, R, modulus->orig_modulus); /* FIXME: can we avoid this? */
     }
-  else if (modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_REDC)
     {
       REDC (R, S, modulus->temp1, modulus);
     }
@@ -958,12 +1124,12 @@ mpres_get_z (mpz_t R, mpres_t S, mpmod_t modulus)
 void 
 mpres_set_ui (mpres_t R, unsigned int n, mpmod_t modulus)
 {
-  if (modulus->repr == MOD_PLAIN || modulus->repr == MOD_BASE2)
+  if (modulus->repr == ECM_MOD_MPZ || modulus->repr == ECM_MOD_BASE2)
     {
       mpz_set_ui (R, n);
       mpz_mod (R, R, modulus->orig_modulus);
     }
-  else if (modulus->repr == MOD_MODMULN || modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_MODMULN || modulus->repr == ECM_MOD_REDC)
     {
       mpz_set_ui (modulus->temp1, n);
       mpz_mul_2exp (modulus->temp1, modulus->temp1, modulus->bits);
@@ -985,13 +1151,13 @@ int
 mpres_invert (mpres_t R, mpres_t S, mpmod_t modulus)
 {
   ASSERT_NORMALIZED (S);
-  if (modulus->repr == MOD_PLAIN || modulus->repr == MOD_BASE2)
+  if (modulus->repr == ECM_MOD_MPZ || modulus->repr == ECM_MOD_BASE2)
     {
       int res = mpz_invert (R, S, modulus->orig_modulus);
       ASSERT_NORMALIZED (R);
       return res;
     }
-  else if (modulus->repr == MOD_MODMULN)
+  else if (modulus->repr == ECM_MOD_MODMULN)
     {
       if (mpz_invert (modulus->temp2, S, modulus->orig_modulus))
         {
@@ -1003,7 +1169,7 @@ mpres_invert (mpres_t R, mpres_t S, mpmod_t modulus)
       else
         return 0;
     }
-  else if (modulus->repr == MOD_REDC)
+  else if (modulus->repr == ECM_MOD_REDC)
     {
       MPZ_NORMALIZED (S);
       if (mpz_invert (modulus->temp2, S, modulus->orig_modulus))
