@@ -20,13 +20,19 @@
   MA 02111-1307, USA.
 */
 
-#include <stdlib.h>
 #include <stdio.h>
-#include <limits.h> /* for UINT_MAX */
-#include <gmp.h>
+#include <stdlib.h> /* for abs if assertions enabled */
 #include "ecm-impl.h"
 #include "ecm-gmp.h"
 
+#if HAVE_LIMITS_H
+# include <limits.h>
+#else
+# ifndef UINT_MAX
+#  define UINT_MAX (~(unsigned int) 0)
+# endif
+#endif
+
 /*
 #define DEBUG 1
 #define CHECKSUM 1
@@ -37,8 +43,6 @@ static int gt_inited = 0;
 static int radix2 = 0;
 unsigned int Fermat;
 
-#define min(a,b) ((a)<(b)?(a):(b))
-
 #define CACHESIZE 512U
 
 /* a' <- a+b, b' <- a-b. */
@@ -49,10 +53,10 @@ unsigned int Fermat;
   F_mod_gt (b, n);    \
   F_mod_1 (a, n);
 
-mp_limb_t __gmpn_mod_34lsub1 (mp_limb_t*, mp_size_t);
+__GMP_DECLSPEC mp_limb_t __gmpn_mod_34lsub1 (mp_limb_t*, mp_size_t);
 
 /* compute remainder modulo 2^(mp_bits_per_limb*3/4)-1 */
-#ifndef HAVE_NATIVE_mpn_mod_34lsub1
+#ifndef HAVE___GMPN_MOD_34LSUB1
 mp_limb_t
 __gmpn_mod_34lsub1 (mp_limb_t *src, mp_size_t size)
 {
@@ -60,7 +64,7 @@ __gmpn_mod_34lsub1 (mp_limb_t *src, mp_size_t size)
   mp_limb_t r, d;
 
   ASSERT(BITS_PER_MP_LIMB % 4 == 0);
-  tp = malloc (size * sizeof(mp_limb_t));
+  tp = malloc (size * sizeof (mp_limb_t));
   MPN_COPY (tp, src, size);
   d = (mp_limb_t) 1 << (3 * (mp_bits_per_limb / 4)) - (mp_limb_t) 1;
   mpn_divmod_1 (&r, tp, size, d);
@@ -71,7 +75,7 @@ __gmpn_mod_34lsub1 (mp_limb_t *src, mp_size_t size)
 
 /* RS -> RS (mod 2^n+1). If input |RS| < 2^(2*n), result |RS| < 2^(n+1) */
 
-static INLINE void 
+static inline void 
 F_mod_1 (mpz_t RS, unsigned int n)
 {
   mp_size_t size;
@@ -102,7 +106,7 @@ F_mod_1 (mpz_t RS, unsigned int n)
 
 /* R = gt (mod 2^n+1) */
 
-static INLINE void 
+static inline void 
 F_mod_gt (mpz_t R, unsigned int n)
 {
   mp_size_t size;
@@ -140,7 +144,7 @@ F_mod_gt (mpz_t R, unsigned int n)
 static void 
 F_mulmod (mpz_t R, mpz_t S1, mpz_t S2, unsigned int n)
 {
-  mp_size_t n2 = n / __GMP_BITS_PER_MP_LIMB;
+  int n2 = n / __GMP_BITS_PER_MP_LIMB; /* type of _mp_size is int */
   F_mod_1 (S1, n);
   F_mod_1 (S2, n);
   while (mpz_size (S1) > (unsigned) n2)
@@ -157,7 +161,7 @@ F_mulmod (mpz_t R, mpz_t S1, mpz_t S2, unsigned int n)
                "Trying again\n", n, (unsigned long) mpz_sizeinbase (S2, 2));
       F_mod_1 (S2, n);
     }
-#if defined(HAVE_GWNUM)
+#if defined(HAVE_GWNUM) && !defined (TUNE)
   if (n >= GWTHRESHOLD)
     {
 #ifdef DEBUG
@@ -188,10 +192,12 @@ F_mulmod (mpz_t R, mpz_t S1, mpz_t S2, unsigned int n)
 #endif
       return;
     }
-#elif defined(HAVE_FFT)
+#elif defined(HAVE___GMPN_MUL_FFT)
   if (n >= 32768)
     {
       unsigned long k;
+      
+      /* WARNING: _mpz_realloc does not keep the value!!! */
       _mpz_realloc (gt, n2 + 1);
       k = mpn_fft_best_k (n2, S1 == S2);
       mpn_mul_fft (PTR(gt), n2, PTR(S1), ABSIZ(S1), PTR(S2), ABSIZ(S2), k);
@@ -221,6 +227,7 @@ mpz_absadd_2exp (mpz_t RS, unsigned int e)
   sgn = (mpz_sgn (RS) >= 0) ? 1 : -1;
   
   if (limb_idx >= RS->_mp_alloc)
+    /* WARNING: mpz_realloc2 does not keep the value!!! */
     mpz_realloc2 (RS, (limb_idx + 1) * GMP_NUMB_BITS);
   
   /* Now RS->_mp_alloc > limb_idx) */
@@ -238,6 +245,7 @@ mpz_absadd_2exp (mpz_t RS, unsigned int e)
   if (cy)
     {
       if (RS->_mp_alloc <= siz)
+        /* WARNING: mpz_realloc2 does not keep the value!!! */
         mpz_realloc2 (RS, (siz + 1) * GMP_NUMB_BITS);
 
       RS->_mp_d[siz] = 1;
@@ -1442,3 +1450,10 @@ F_mul_trans (mpz_t *R, mpz_t *A, mpz_t *B, unsigned int len, unsigned int n,
   
   return r;
 }
+
+void F_clear ()
+{
+  if (gt_inited)
+    mpz_clear (gt);
+  gt_inited = 0;
+}
